package com.contextspace;

import static org.junit.Assert.*;

import java.io.Serializable;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.annotations.Index;
import javax.persistence.Basic;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Embeddable;
import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.ManyToOne;
import javax.persistence.MappedSuperclass;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.Persistence;
import javax.persistence.TableGenerator;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;
import javax.persistence.Transient;
import javax.persistence.TypedQuery;
import javax.persistence.Version;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlTransient;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;



public class AddressTest {

    @Test
    public void test() {
        EntityManagerFactory emf = Persistence.createEntityManagerFactory("objectdb:$objectdb/db/test.tmp;drop");
        EntityManager em = emf.createEntityManager();
     
        
        em.getTransaction().begin();

        CDI cdi = new CDI("jbloggs","Joe", "Bloggs", new Date());
        em.persist(cdi);
     
        em.getTransaction().commit();
     
        TypedQuery<Long> query = em.createQuery("SELECT COUNT($1) FROM CDI $1 WHERE $1.loginId LIKE '%' ESCAPE '\\'", Long.class);
        Long cdiCount = query.getSingleResult();
        assertEquals(1, cdiCount.intValue());
        
        TypedQuery<CDI> query2 = em.createQuery("SELECT c FROM CDI c", CDI.class);
        List<CDI> results = query2.getResultList();
        assertEquals(1, results.size());
        
        System.out.println(printPrettyXml(results.get(0).toString(),4));
        
        Address address = new Address();
        address.setAddressline("Pacific Hwy");
        address.setCity("Sydney");
        address.setPostalcode("2066");
        address.setPrimary(true);
        address.setType("home");
        cdi.addAddress(address);
        
        em.getTransaction().begin();
        em.merge(cdi);
        em.getTransaction().commit();
        
        TypedQuery<CDI> query3 = em.createQuery("SELECT c FROM CDI c", CDI.class);
        List<CDI> results2 = query3.getResultList();
        assertEquals(1, results2.size());
        
        System.out.println(printPrettyXml(results2.get(0).toString(),4));
        
        em.close();
        emf.close();
    }

    
    @Entity
    @TableGenerator(name="IdTableGen", initialValue=1, allocationSize=50)
    public static class CDI implements Serializable{
        private static final long serialVersionUID = 1L;
       
        private Date creationDate = null;
        private String loginId = null;
        private String firstname = null;
        private String lastname = null;
        
        private ContactDetails contactDetails = new ContactDetails();
        
        private long id;
        
        /**
         * Default Constructor
         */
        public CDI() {
            
        }
        
        /**
         * Constructor
         * @param firstname
         * @param lastname
         */
        public CDI(String firstname, String lastname) {
            this.firstname = firstname;
            this.lastname = lastname;
            this.creationDate = new Date();
        }
        
        /**
         * Constructor
         * @param firstname
         * @param lastname
         * @param creationDate
         */
        public CDI(String loginId, String firstname, String lastname, Date creationDate) {
            this.loginId = loginId;
            this.firstname = firstname;
            this.lastname = lastname;
            this.creationDate = creationDate;
        }
            
        /**
         * @return the creationDate
         */
        @Temporal(TemporalType.TIMESTAMP)
        public Date getCreationDate() {
            return creationDate;
        }
        
        /**
         * @param creationDate the creationDate to set
         */
        public void setCreationDate(Date creationDate) {
            this.creationDate = creationDate;
        }
        
        /**
         * @return the object id
         */
        @Id
        @GeneratedValue(strategy=GenerationType.TABLE, generator="IdTableGen")
        public long getId() {
            return id;
        }

        /**
         * @param id the id to be set
         */
        public void setId(long id) {
            this.id = id;
        }

        /**
         * @return the firstname
         */
        @Index(unique="false")
        public String getFirstname() {
            return firstname;
        }

        /**
         * @param firstname the firstname to set
         */
        public void setFirstname(String firstname) {
            this.firstname = firstname;
        }

        /**
         * @return the lastname
         */
        @Index(unique="false")
        public String getLastname() {
            return lastname;
        }

        /**
         * @param lastname the lastname to set
         */
        public void setLastname(String lastname) {
            this.lastname = lastname;
        }

        /**
         * @return the loginId
         */
        @Index(unique="true")
        @Column(nullable=false,unique=true)
        public String getLoginId() {
            return loginId;
        }

        /**
         * @param loginId the loginId to set
         */
        public void setLoginId(String loginId) {
            this.loginId = loginId;
        }

        public void addAddress(Address address) {
            contactDetails.addAddress(address);
        }
        
        public void removeAddress(Address address) {
            contactDetails.removeAddress(address);
        }
        
        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        @Override
        public String toString() {
            StringBuffer buf = new StringBuffer();
            
            buf.append("<cdi");
            if(getId() > 0) {
                buf.append(" id=\"").append(getId()).append("\"");
            }
            buf.append(">");
     
            if(firstname != null)
                buf.append("<firstname>").append(firstname).append("</firstname>");
            if(lastname != null)
                buf.append("<lastname>").append(lastname).append("</lastname>");
            if(loginId != null)
                buf.append("<loginid>").append(loginId).append("</loginid>");
            
            buf.append(contactDetails.toString());
            
            buf.append("</cdi>");
            return buf.toString();
        }

        /**
         * @return the contactDetails
         */
        @Embedded
        public ContactDetails getContactDetails() {
            return contactDetails;
        }

        /**
         * @param contactDetails the contactDetails to set
         */
        public void setContactDetails(ContactDetails contactDetails) {
            this.contactDetails = contactDetails;
        } 
        
        
    }
    
    @Embeddable
    public static class ContactDetails implements Serializable{
        private static final long serialVersionUID = -8395583852077269007L;

        private List<Address> addresses = new ArrayList<Address>();
        
        /**
         * Default Constructor
         */
        public ContactDetails() {
            
        }
        
        /**
         * @param emailAddresses
         * @param addresses
         * @param phoneNumbers
         */
        public ContactDetails(List<Address> addresses) {
            this.addresses = addresses;
        }

        /**
         * @param address
         */
        public void addAddress(Address address) {
            addresses.add(address);
        }
        
        /**
         * @return the addresses
         */
        @OneToMany(cascade=CascadeType.ALL, fetch=FetchType.EAGER)
        @XmlElementWrapper(name="addresses")
        @XmlElement(name = "address")
        public List<Address> getAddresses() {
            return addresses;
        }

        /**
         * @param addresses the addresses to set
         */
        public void setAddresses(List<Address> addresses) {
            this.addresses = addresses;
        }
        
        /**
         * @param addressToRemove
         */
        public void removeAddress(Address addressToRemove) {
            addresses.remove(addressToRemove);
        }
        
        /**
         * @param addressToReplace
         */
        public void replaceAddress(Address addressToReplace) {
            for(int i=0; i<addresses.size(); i++) {
                if(addresses.get(i).equals(addressToReplace)) {
                    addresses.set(i, addressToReplace);
                }
            }
        }
        
        /**
         * Remove all Addresses
         */
        public void removeAllAddresses() {
            addresses.clear();
        }
        
        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        @Override
        public String toString() {
            StringBuffer buf = new StringBuffer();
            if(addresses != null && !addresses.isEmpty()) {
                buf.append("<contactDetails>");
                if(addresses != null) {
                    for(Iterator<Address> i = addresses.iterator(); i.hasNext();) {
                        Address address = i.next();
                        buf.append(address.toString());
                    }
                }
                buf.append("</contactDetails>");     
            }
            return buf.toString();
        }
        
    
    }

    
    @Entity
    @XmlRootElement(name="address")
    public static class Address extends AbstractContactDetail {
        private static final long serialVersionUID = 1L;

        private String city = null;
        private String state = null;
        private String province = null;
        private String country = null;
        private String postalcode = null;
        
        //This may contain mutliple lines. Each line will be separated by a delimiter
        private String addressline = null;
        
        
        /**
         * Default Constructor
         * Sets primary false
         */
        public Address() {
            this.setPrimary(false);
            this.type = "home"; //default address to home
        }
        
        /**
         * Constructor
         * @param primary
         */
        public Address(boolean primary) {
            this.setPrimary(primary);
            this.type = "home"; //default address to home
        }
        
        /**
         * Constructor taking primary and the type of this address
         * @param primary
         */
        public Address(boolean primary, String type) {
            this.setPrimary(primary);
            this.type = type;
        }
        
        
        /* (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        @Override
        public String toString() {
            StringBuffer buf = new StringBuffer();
            buf.append("<address");
            buf.append(super.toString());
            
            if(addressline != null) {
                buf.append("<addressline>").append(addressline).append("</addressline>\n");      
            }
            if(province != null) {
                buf.append("<province>").append(province).append("</province>\n");      
            }
            if(state != null) {
                buf.append("<state>").append(state).append("</state>\n");      
            }
            if(city != null) {
                buf.append("<city>").append(city).append("</city>\n");      
            }
            if(country != null) {
                buf.append("<country>").append(country).append("</country>\n");      
            }
            if(postalcode != null) {
                buf.append("<postalcode>").append(postalcode).append("</postalcode>\n");      
            }   
            buf.append("</address>");
            return buf.toString();
        }

        /* (non-Javadoc)
         * @see java.lang.Object#equals(java.lang.Object)
         */
        @Override
        public boolean equals(Object obj) {
            if(!(obj instanceof Address)) {
                return false;
            }
            return super.equals(obj);
        }

        /**
         * @return the city
         */
        public String getCity() {
            return city;
        }

        /**
         * @param city the city to set
         */
        public void setCity(String city) {
            this.city = city;
        }

        /**
         * @return the state
         */
        public String getState() {
            return state;
        }

        /**
         * @param state the state to set
         */
        public void setState(String state) {
            this.state = state;
        }

        /**
         * @return the province
         */
        public String getProvince() {
            return province;
        }

        /**
         * @param province the province to set
         */
        public void setProvince(String province) {
            this.province = province;
        }

        /**
         * @return the country
         */
        public String getCountry() {
            return country;
        }

        /**
         * @param country the country to set
         */
        public void setCountry(String country) {
            this.country = country;
        }

        /**
         * @return the postalcode
         */
        public String getPostalcode() {
            return postalcode;
        }

        /**
         * @param postalcode the postalcode to set
         */
        public void setPostalcode(String postalcode) {
            this.postalcode = postalcode;
        }

        /**
         * @return the addressline
         */
        public String getAddressline() {
            return addressline;
        }

        /**
         * @param addressline the addressline to set
         */
        public void setAddressline(String addressline) {
            this.addressline = addressline;
        }
    }

    @MappedSuperclass
    public static abstract class AbstractContactDetail {

      private static final long serialVersionUID = 1L;
      private boolean primary = false;
      private BasicMetaData basicMetaData = new BasicMetaData();
      
      protected String type;
      
      public AbstractContactDetail(){
          super();
          
          System.out.println("AbstractDetail Construct: " + (basicMetaData == null ? "null" : " not null"));
      }
      
      /**
       * @return the primary
       */
      @Column(name="isPrimary")
      @XmlAttribute(name="primary")
      public boolean getPrimary() {
          return primary;
      }
      
      /**
       * @param primary the primary to set
       */
      public void setPrimary(boolean primary) {
          this.primary = primary;
          
          setObjectMetaDataChange();
      }
      
      /**
       * @return the basicMetaData
       */
      @Embedded
      @XmlElement(name="basicMetaData")
      public BasicMetaData getBasicMetaData() {
          return basicMetaData;
      }

      /**
       * @param basicMetaData the basicMetaData to set
       */
      public void setBasicMetaData(BasicMetaData basicMetaData) {
          this.basicMetaData = basicMetaData;
      }
      
      /**
       * set object meta data change
       */
      protected void setObjectMetaDataChange() {
          // Set the metadata
          if(basicMetaData == null) {
              
              System.out.println("metadata is null: superclass: " + this.getClass().getSuperclass().getName());
          }
          
          basicMetaData.setLastModifiedBy("me");
          basicMetaData.setLastModifiedDate(new Date());
      }
      
      /**
       * @return the type
       */
      @Index
      public String getType() {
          return type;
      }

      /**
       * @param type the type to set
       */
      public void setType(String type) {
          this.type = type;
          
          setObjectMetaDataChange();
      }
      
      /* (non-Javadoc)
       * @see java.lang.Object#toString()
       */
      @Override
      public String toString() {
          StringBuffer buf = new StringBuffer();
          
          buf.append(" type=\"").append(type).append("\"");
          buf.append(" primary=\"").append(primary).append("\"");
          buf.append(">");
          buf.append(super.toString());
          return buf.toString();
      }
      
      @Override
      public boolean equals(Object obj) {
          if(!(obj instanceof AbstractContactDetail)) {
              return false;
          }      
          AbstractContactDetail other = (AbstractContactDetail)obj;
          if(!super.equals(obj)){
              return false;
          }
          
          if(getPrimary() != other.getPrimary() ){
              return false;
          }
          if((getType()==null && other.getType()!=null) || (getType()!=null&&!getType().equals(other.getType())) ){
              return false;
          }
          
          return true;
      }
  }

  
  


  @Embeddable
  public static class BasicMetaData extends AbstractMetaData {
      
      /**
       * Basic Constructor
       */
      public BasicMetaData() {
          super();
      }
  }
  
  @MappedSuperclass
  public static abstract class AbstractMetaData implements Serializable{

      private Date creationDate = null;
      private Date lastModifiedDate = null;
      private String lastModifiedBy = null;

      /**
       * Default Constructor
       * Set creation date
       */
      public AbstractMetaData() {
          super();
          this.creationDate = new Date();
      }
      
      /**
       * @return the creationDate
       */
      @Temporal(TemporalType.TIMESTAMP)
      @XmlAttribute(name = "creationDate")
      public Date getCreationDate() {
          return creationDate;
      }
      
      
      /**
       * @param creationDate the creationDate to set
       */
      public void setCreationDate(Date creationDate) {
          this.creationDate = creationDate;
      }
      
      /**
       * @return the lastModifiedDate
       */
      @Temporal(TemporalType.TIMESTAMP)
      @XmlAttribute(name = "lastModifiedDate")
      public Date getLastModifiedDate() {
          return lastModifiedDate;
      }
      /**
       * @param lastModifiedDate the lastModifiedDate to set
       */
      public void setLastModifiedDate(Date lastModifiedDate) {
          this.lastModifiedDate = lastModifiedDate;
      }
      
      /**
       * @return the lastModifiedBy
       */
      @XmlAttribute(name = "lastModifiedBy")
      public String getLastModifiedBy() {
          return lastModifiedBy;
      }
      /**
       * @param lastModifiedBy the lastModifiedBy to set
       */
      public void setLastModifiedBy(String lastModifiedBy) {
          this.lastModifiedBy = lastModifiedBy;
      }
      
  }

  
      
      
  
    
    /**
     * Print pretty xml
     * @param input
     * @param indent
     * @return
     * @throws TransformerException
     */
    public static String printPrettyXml(String input, int indent) {
        try {
            Source xmlInput = new StreamSource(new StringReader(input));
            StringWriter stringWriter = new StringWriter();
            StreamResult xmlOutput = new StreamResult(stringWriter);
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer();
            
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", String.valueOf(indent));
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.transform(xmlInput, xmlOutput);
            return xmlOutput.getWriter().toString();
        }
        catch(TransformerException e) {
            return "Exception thrown during transform of xml: " + e.getMessage();
        }
    }
}
